use std::collections::HashMap;
use std::panic::AssertUnwindSafe;
use std::process::{exit, self};
use std::sync::Arc;
use std::sync::atomic::{AtomicBool, Ordering};
use std::thread;
use std::time::{SystemTime, Duration};

use indicatif::{ProgressBar, ProgressStyle};
use structopt::StructOpt;

use fluvio::Fluvio;
use fluvio_test_util::test_meta::{BaseCli, TestCase, TestCli, TestOption};
use fluvio_test_util::test_meta::test_result::TestResult;
use fluvio_test_util::test_meta::environment::{EnvDetail, EnvironmentSetup};
use fluvio_test_util::setup::TestCluster;
use fluvio_test_util::test_runner::test_driver::{TestDriver};
use fluvio_test_util::test_runner::test_meta::FluvioTestMeta;
use fluvio_test_util::{async_process};

// This is important for `inventory` crate
#[allow(unused_imports)]
use fluvio_test::tests as _;
use sysinfo::{System, SystemExt, get_current_pid, ProcessExt, Signal, Process, Pid};
use tracing::debug;

//const CI_FAIL_FLAG: &str = "/tmp/CI_FLUVIO_TEST_FAIL";

fn main() {
    let option = BaseCli::from_args();

    debug!("{:?}", option);

    let test_name = option.environment.test_name.clone();

    // Get test from inventory
    let test_meta =
        FluvioTestMeta::from_name(&test_name).expect("StructOpt should have caught this error");

    let mut subcommand = vec![test_name.clone()];

    if let Some(TestCli::Args(args)) = option.test_cmd_args {
        // Add the args to the subcommand
        subcommand.extend(args);
    }

    // We want to get a TestOption compatible struct back
    let test_opt: Box<dyn TestOption> = (test_meta.validate_fn)(subcommand);

    println!("Start running fluvio test runner");
    fluvio_future::subscriber::init_logger();

    // Test connecting to a cluster
    // Deploy a cluster if requested
    cluster_setup(&option.environment).expect("Failed to connect to a cluster");

    // Check on test requirements before running the test
    if !TestDriver::is_env_acceptable(
        &(test_meta.requirements)(),
        &TestCase::new(option.environment.clone(), test_opt.clone()),
    ) {
        exit(-1);
    }

    let test_result = run_test(option.environment.clone(), test_opt, test_meta);

    cluster_cleanup(option.environment);
    println!("{}", test_result);

    if test_result.success {
        exit(0)
    } else {
        exit(-1)
    }
}

fn run_test(
    environment: EnvironmentSetup,
    test_opt: Box<dyn TestOption>,
    test_meta: &FluvioTestMeta,
) -> TestResult {
    let start = SystemTime::now();
    let test_case = TestCase::new(environment.clone(), test_opt);
    let test_cluster_opts = TestCluster::new(environment.clone());
    let test_driver = TestDriver::new(Some(test_cluster_opts));

    let ok_signal = Arc::new(AtomicBool::new(false));
    let fail_signal = Arc::new(AtomicBool::new(false));
    signal_hook::flag::register(signal_hook::consts::SIGUSR1, Arc::clone(&ok_signal))
        .expect("fail to register ok signal hook");
    signal_hook::flag::register(signal_hook::consts::SIGUSR2, Arc::clone(&fail_signal))
        .expect("fail to register fail signal hook");

    // println!("supported signals: {:?}", System::SUPPORTED_SIGNALS);
    let root_pid = get_current_pid().expect("Unable to get current pid");
    debug!(?root_pid, "current root pid");
    let mut sys = System::new();
    if !sys.refresh_process(root_pid) {
        panic!("Unable to refresh root");
    }
    let root_process = sys.process(root_pid).expect("Unable to get root process");
    let _child_pid = match fork::fork() {
        Ok(fork::Fork::Parent(child_pid)) => child_pid,
        Ok(fork::Fork::Child) => {
            println!("starting test in child process");
            // put panic handler, this shows proper stack trace in the console unlike hook
            let status = std::panic::catch_unwind(AssertUnwindSafe(|| {
                (test_meta.test_fn)(test_driver, test_case)
            }));
            match status {
                Ok(_) => {
                    println!("test passed, signalling success to parent");
                    root_process.kill_with(Signal::User1);
                    process::exit(0)
                }
                Err(err) => {
                    println!("test failed {:#?}, signalling parent", err);
                    root_process.kill_with(Signal::User2);
                    process::exit(1);
                }
            };
        }
        Err(_) => panic!("Fork failed"),
    };

    debug!("running waiting for signal");

    const TICK: Duration = Duration::from_millis(100);
    let disable_timeout = environment.disable_timeout;
    let timeout = environment.timeout();
    let mut message_spinner = create_spinning_indicator();
    let mut timed_out = false;

    loop {
        if ok_signal.load(Ordering::Relaxed) || fail_signal.load(Ordering::Relaxed) {
            debug!("signal received");
            break;
        }

        let elapsed = start.elapsed().expect("Unable to get elapsed time");
        if !disable_timeout && elapsed > timeout {
            timed_out = true;
            break;
        }
        thread::sleep(TICK);
        if let Some(pb) = &mut message_spinner {
            pb.set_message(format!("waiting for test {} seconds", elapsed.as_secs()));
        }
    }

    let ok = ok_signal.load(Ordering::Relaxed);
    let fail = fail_signal.load(Ordering::Relaxed);

    debug!(ok, fail, "signal status");

    if timed_out {
        println!("Test timed out after {} seconds", timeout.as_secs());
        kill_child_processes(root_process);
        TestResult {
            success: true,
            duration: start.elapsed().unwrap(),
            ..std::default::Default::default()
        }
    } else if ok {
        println!("Test passed");
        TestResult {
            success: true,
            duration: start.elapsed().unwrap(),
            ..std::default::Default::default()
        }
    } else {
        println!("test failed, killing all child processes");
        kill_child_processes(root_process);
        TestResult {
            success: false,
            duration: start.elapsed().unwrap(),
            ..std::default::Default::default()
        }
    }
}

/// kill all children of the root processes
fn kill_child_processes(root_process: &Process) {
    let root_pid = root_process.pid();
    let mut sys2 = System::new();
    sys2.refresh_processes();
    let g_id = root_process.gid;

    let proceses = sys2.processes();

    fn is_root(process: &Process, r_id: Pid, processes: &HashMap<Pid, Process>) -> bool {
        if let Some(parent_id) = process.parent() {
            if parent_id == r_id {
                true
            } else if let Some(parent_process) = processes.get(&parent_id) {
                is_root(parent_process, r_id, processes)
            } else {
                false
            }
        } else {
            false
        }
    }

    for (pid, process) in proceses {
        if pid != &root_pid && process.gid == g_id && is_root(process, root_pid, proceses) {
            println!("killing child test pid {} name {}", pid, process.name());
            process.kill();
        }
    }
}

fn cluster_cleanup(option: EnvironmentSetup) {
    if option.cluster_delete() {
        let mut setup = TestCluster::new(option);

        let cluster_cleanup_wait = async_process!(
            async {
                setup.remove_cluster().await;
            },
            "cluster_cleanup"
        );
        let _ = cluster_cleanup_wait
            .join()
            .expect("Cluster cleanup wait failed");
    }
}

// FIXME: Need to confirm SPU options count match cluster. Offer self-correcting behavior
fn cluster_setup(option: &EnvironmentSetup) -> Result<(), ()> {
    let cluster_setup_wait = async_process!(
        async {
            if option.remove_cluster_before() {
                println!("Deleting existing cluster before starting test");
                let mut setup = TestCluster::new(option.clone());
                setup.remove_cluster().await;
            }

            if option.cluster_start() || option.remove_cluster_before() {
                println!("Starting cluster and testing connection");
                let mut test_cluster = TestCluster::new(option.clone());

                test_cluster
                    .start()
                    .await
                    .expect("Unable to connect to fresh test cluster");
            } else {
                println!("Testing connection to Fluvio cluster in profile");
                Fluvio::connect()
                    .await
                    .expect("Unable to connect to Fluvio test cluster via profile");
            }
        },
        "cluster setup"
    );

    let _ = cluster_setup_wait
        .join()
        .expect("Cluster setup wait failed");

    Ok(())
}

fn create_spinning_indicator() -> Option<ProgressBar> {
    if std::env::var("CI").is_ok() {
        None
    } else {
        let pb = ProgressBar::new(1);
        pb.set_style(
            ProgressStyle::default_bar()
                .template("{msg} {spinner}")
                .tick_chars("/-\\|"),
        );
        pb.enable_steady_tick(100);
        Some(pb)
    }
}

#[cfg(test)]
mod tests {
    // CLI Tests

    use std::time::Duration;
    use structopt::StructOpt;
    use fluvio_test_util::test_meta::{BaseCli, TestCli};
    use fluvio_test_util::test_meta::environment::EnvDetail;
    use fluvio_test::tests::smoke::SmokeTestOption;

    #[test]
    fn valid_test_name() {
        let args = BaseCli::from_iter_safe(vec!["fluvio-test", "smoke"]);

        assert!(args.is_ok());
    }

    #[test]
    fn invalid_test_name() {
        let args = BaseCli::from_iter_safe(vec!["fluvio-test", "testdoesnotexist"]);

        assert!(args.is_err());
    }

    #[test]
    fn extra_vars() {
        let args = BaseCli::from_iter(vec![
            "fluvio-test",
            "smoke",
            "--",
            "--producer-iteration=9000",
            "--producer-record-size=1000",
        ]);

        if let Some(TestCli::Args(cmd)) = args.test_cmd_args {
            // Structopt commands expect an arbitrary binary name before the test args
            let mut subcommand = vec![args.environment.test_name.clone()];
            subcommand.extend(cmd);

            let smoke_test_case = SmokeTestOption::from_iter(subcommand);

            let expected = SmokeTestOption {
                producer_iteration: 9000,
                producer_record_size: 1000,
                ..Default::default()
            };

            assert_eq!(smoke_test_case, expected);
        } else {
            panic!("test args not parsed")
        }
    }

    #[test]
    fn topic() {
        let args = BaseCli::from_iter(vec![
            "fluvio-test",
            "smoke",
            "--topic-name",
            "not_the_default_topic_name",
        ]);

        assert_eq!(
            args.environment.base_topic_name(),
            "not_the_default_topic_name"
        );
    }

    #[test]
    fn spu() {
        let args = BaseCli::from_iter(vec!["fluvio-test", "smoke", "--spu", "5"]);

        assert_eq!(args.environment.spu, 5);
    }

    #[test]
    fn timeout() {
        let args = BaseCli::from_iter(vec!["fluvio-test", "smoke", "--timeout", "9000"]);

        assert_eq!(args.environment.timeout(), Duration::from_secs(9000));
    }

    //// We validate that the behavior of cluster_setup and cluster_cleanup work as expected
    //// The clusters are the same if cluster addr from the first run is the same as the second run
    ////
    //// You may need to run `make minikube_image` if this fails with:
    //// 'Failed to install k8 cluster: InstallK8(HelmChartNotFound("fluvio/fluvio-app:a.b.c-xyz"))'
    //// This won't work with k8 until we fix: https://github.com/infinyon/fluvio/issues/859
    ////
    //// Local cluster does not work in `cargo test`:
    //// 'Failed to install local cluster: InstallLocal(IoError(Os { code: 2, kind: NotFound, message: "No such file or directory" }))'
    //#[test]
    //#[ignore]
    //fn skip_cluster_delete_then_skip_cluster_start() {
    //    use super::*;
    //    use fluvio::config::ConfigFile;
    //    use fluvio_future::task::run_block_on;

    //    run_block_on(async {
    //        let skip_cluster_delete_cmd = CliArgs::from_iter(vec![
    //            "fluvio-test",
    //            "smoke",
    //            "--keep-cluster",
    //            "--local",
    //            "--develop",
    //            "--skip-checks",
    //        ]);
    //        let test_case: TestCase = skip_cluster_delete_cmd.into();
    //        let _fluvio_client = cluster_setup(&test_case).await;
    //        let cluster_addr_1 = {
    //            let config_file = ConfigFile::load_default_or_new().expect("Default config");
    //            let fluvio_config = config_file.config().current_cluster().unwrap();
    //            fluvio_config.endpoint.clone()
    //        };
    //        cluster_cleanup(test_case).await;

    //        let skip_cluster_start_cmd = CliArgs::from_iter(vec![
    //            "fluvio-test",
    //            "smoke",
    //            "--disable-install",
    //            "--local",
    //            "--develop",
    //            "--skip-checks",
    //        ]);
    //        let test_case: TestCase = skip_cluster_start_cmd.into();
    //        let _fluvio_client = cluster_setup(&test_case).await;
    //        let cluster_addr_2 = {
    //            let config_file = ConfigFile::load_default_or_new().expect("Default config");
    //            let fluvio_config = config_file.config().current_cluster().unwrap();
    //            fluvio_config.endpoint.clone()
    //        };
    //        cluster_cleanup(test_case).await;

    //        assert_eq!(cluster_addr_1, cluster_addr_2);
    //    });
    //}
}
